"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var fs_1 = require("fs");
var net_1 = require("net");
var atvik_1 = require("atvik");
var debug_1 = __importDefault(require("debug"));
var proper_lockfile_1 = require("proper-lockfile");
/**
 * Resolve the given path to a path that can be used with an IPC socket.
 */
function toSocketPath(p) {
    var resolved = path_1.default.resolve(p);
    if (process.platform === 'win32') {
        // On Windows IPC sockets need to be either in \\?\pipe\ or \\.\pipe\
        return path_1.default.join('\\\\?\\pipe', resolved);
    }
    else {
        // Other platforms are assumed to be Unix-like where sockets can be created anywhere
        return resolved;
    }
}
/**
 * Generate a random retry time between 30 and 130 ms.
 */
function randomRetryTime() {
    return Math.floor(30 + Math.random() * 100);
}
var client = Symbol('client');
var server = Symbol('server');
var releaseLock = Symbol('releaseLock');
var readyEvent = Symbol('readyEvent');
var errorEvent = Symbol('errorEvent');
var leaderEvent = Symbol('leaderEvent');
var connectEvent = Symbol('connectEvent');
var connectionEvent = Symbol('connectionEvent');
/**
 * Low-level network that provides sockets to and from a leader. When an
 * instance of this network is connected it can either become the leader, in
 * which case a server socket will be opened, or a client in which it will
 * connect to the leader.
 *
 * When connected the network will attempt to stay connected.
 */
var LowLevelNetwork = /** @class */ (function () {
    function LowLevelNetwork(options) {
        if (typeof options !== 'object') {
            throw new Error('Options are needed to create a new network');
        }
        if (typeof options.path !== 'string') {
            throw new Error('Path for socket is required');
        }
        this.options = {
            path: options.path
        };
        this.debug = debug_1.default('local-machine-network:' + (options.id || ('[' + options.path + ']')));
        this.leader = false;
        this.started = false;
        this[readyEvent] = new atvik_1.Event(this);
        this[errorEvent] = new atvik_1.Event(this);
        this[connectEvent] = new atvik_1.Event(this);
        this[connectionEvent] = new atvik_1.Event(this);
        this[leaderEvent] = new atvik_1.Event(this);
    }
    Object.defineProperty(LowLevelNetwork.prototype, "onConnect", {
        get: function () {
            return this[connectEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LowLevelNetwork.prototype, "onConnection", {
        get: function () {
            return this[connectionEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LowLevelNetwork.prototype, "onLeader", {
        get: function () {
            return this[leaderEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LowLevelNetwork.prototype, "onReady", {
        get: function () {
            return this[readyEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LowLevelNetwork.prototype, "onError", {
        get: function () {
            return this[errorEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Start this network.
     */
    LowLevelNetwork.prototype.start = function () {
        var _this = this;
        if (this.started) {
            // TODO: If we are started return promise that resolves if connected
        }
        this.started = true;
        return new Promise(function (resolve, reject) {
            var readyHandle = _this.onReady(function () {
                readyHandle.unsubscribe();
                errorHandle.unsubscribe();
                resolve();
            });
            var errorHandle = _this.onError(function (err) {
                readyHandle.unsubscribe();
                errorHandle.unsubscribe();
                reject();
            });
            // Attempt to connect or bind
            _this.attemptConnectOrBind();
        });
    };
    /**
     * Attempt to connect or become the master.
     *
     * The flow here is:
     *
     * 1) Attempt to acquire lock
     * 2a) If lock acquired: Bind and become the master
     * 2b): Otherwise: Connect to the master
     */
    LowLevelNetwork.prototype.attemptConnectOrBind = function () {
        var _this = this;
        // Resolve the paths to work with
        var socketPath = toSocketPath(this.options.path);
        // Reset leader flag
        this.leader = false;
        /*
         * Attempt to get a lock, hopefully allowing us to create the
         * socket and become the leader.
         */
        proper_lockfile_1.lock(this.options.path, {
            /*
             * As the socket doesn't exist before we lock we don't use realpath
             * and instead let lock be created wherever.
             */
            realpath: false,
            onCompromised: function () {
                var release = _this[releaseLock];
                if (release) {
                    _this.debug('Lock on server path compromised, attempting release');
                    release()
                        .catch(function (err) { return _this.debug('Unable to unlock, error: ', err); })
                        .then(function () {
                        _this[releaseLock] = undefined;
                        _this.reconnect('Attempting reconnect after compromised lock');
                    });
                }
                else {
                    _this.reconnect('Lock on server path compromised');
                }
            }
        })
            .then(function (release) {
            /**
             * We now have the lock, try creating a server.
             */
            // Store a reference to be able to release the lock
            _this[releaseLock] = release;
            _this.attemptBind(socketPath);
        })
            .catch(function () {
            /*
             * Someone else is already holding the lock - try to
             * connect to them.
             */
            _this.attemptConnect(socketPath);
        });
    };
    /**
     * Unsafe: Release the lock if we have it. This is done if the server
     * fails to bind or if the network is disconnected.
     */
    LowLevelNetwork.prototype._releaseLock = function () {
        var release = this[releaseLock];
        if (!release)
            return Promise.resolve();
        this[releaseLock] = undefined;
        return release();
    };
    LowLevelNetwork.prototype.reconnect = function (message) {
        var _this = this;
        // Check that this network is started
        if (!this.started)
            return;
        // Resolve when to next retry a connection
        var retryTime = randomRetryTime();
        this.debug(message + '; Retrying in ' + retryTime + ' ms');
        setTimeout(function () { return _this.attemptConnectOrBind(); }, retryTime);
    };
    LowLevelNetwork.prototype.attemptConnect = function (socketPath) {
        var _this = this;
        this.debug('Connecting as client');
        var socket = net_1.connect(socketPath, function () {
            _this.debug('Connected to leader');
            _this[client] = socket;
            _this[connectEvent].emit(socket);
            // Indicate that the network is ready
            _this[readyEvent].emit();
        });
        socket.on('error', function (err) {
            // Mark the client as non-existent
            var hadConnected = !!_this[client];
            _this[client] = undefined;
            // Resolve a readable debug message
            var message;
            var code = typeof err.code !== 'undefined' ? err.code : 'Unknown Error';
            if (hadConnected) {
                message = 'Disconnected from server with ' + code;
            }
            else {
                message = 'Could not connect to server, got code ' + code;
            }
            _this.reconnect(message);
        });
        socket.on('close', function (hadError) {
            if (hadError)
                return;
            // Mark the client as non-existent
            var hadConnected = !!_this[client];
            _this[client] = undefined;
            // Request a reconnect
            _this.reconnect('Disconnected from server without error');
        });
    };
    LowLevelNetwork.prototype.attemptBind = function (socketPath, attempt) {
        var _this = this;
        if (attempt === void 0) { attempt = 0; }
        if (attempt > 10) {
            // Already tried to bind 10 times - just give up at this point
            this.debug('Could not bind server socket, giving up after 10 tries');
            this[errorEvent].emit(new Error('Could not bind server socket, giving up after 10 tries'));
            return;
        }
        this.debug('Creating socket as leader');
        // Try to unlink the socket path
        fs_1.unlink(socketPath, function (unlinkErr) {
            if (unlinkErr && unlinkErr.code !== 'ENOENT') {
                var retryTime = randomRetryTime();
                _this.debug('Could not unlink existing socket, errored with: ' + unlinkErr.code + '; Retrying in', retryTime, 'ms');
                setTimeout(function () { return _this.attemptBind(socketPath, attempt + 1); }, retryTime);
                return;
            }
            var serverSocket = net_1.createServer();
            serverSocket.listen(socketPath, function () {
                // Server was successfully bound, store reference and resolve callback
                _this.debug('Hosting network as leader');
                _this[server] = serverSocket;
                // Mark as leader and emit event
                _this.leader = true;
                _this[leaderEvent].emit(serverSocket);
                // Indicate that the network is ready
                _this[readyEvent].emit();
            });
            serverSocket.on('connection', function (socket) {
                _this[connectionEvent].emit(socket);
            });
            serverSocket.on('error', function (err) {
                // Something went wrong, close this server and reconnect
                // Mark the server as non-existent
                var hadBound = !!_this[server];
                _this[server] = undefined;
                if (hadBound) {
                    // Make sure the socket is closed
                    serverSocket.close();
                    _this.reconnect('Server errored with ' + err.code);
                }
                else {
                    var retryTime = randomRetryTime();
                    _this.debug('Could not bind socket, errored with ' + err.code + '; Retrying in', retryTime, 'ms');
                    setTimeout(function () { return _this.attemptBind(socketPath, attempt + 1); }, retryTime);
                }
            });
        });
    };
    LowLevelNetwork.prototype.stop = function () {
        var _this = this;
        this.started = false;
        var serverSocket = this[server];
        if (serverSocket) {
            return new Promise(function (resolve, reject) {
                serverSocket.close(function () {
                    // First unlink the socket path
                    var socketPath = toSocketPath(_this.options.path);
                    fs_1.unlink(socketPath, function (err) {
                        _this._releaseLock()
                            .then(resolve)
                            .catch(reject);
                    });
                });
            });
        }
        else {
            var clientSocket_1 = this[client];
            if (clientSocket_1) {
                return new Promise(function (resolve, reject) {
                    clientSocket_1.destroy();
                    resolve();
                });
            }
            else {
                return Promise.resolve();
            }
        }
    };
    return LowLevelNetwork;
}());
exports.LowLevelNetwork = LowLevelNetwork;
//# sourceMappingURL=low-level.js.map