"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var atvik_1 = require("atvik");
var debug_1 = __importDefault(require("debug"));
var low_level_1 = require("./low-level");
var msgpack_lite_1 = __importDefault(require("msgpack-lite"));
var connectEvent = Symbol('connectEvent');
var disconnectEvent = Symbol('disconnectEvent');
var leaderEvent = Symbol('leaderEvent');
var connectionEvent = Symbol('connectionEvent');
var messageEvent = Symbol('messageEvent');
/**
 * Adapter over a low-level network for those cases where clients want to
 * exchange objects with the leader. This implementation will use MessagePack
 * to encode and decode objects.
 */
var ObjectNetwork = /** @class */ (function () {
    /**
     * Create a new object network.
     */
    function ObjectNetwork(options) {
        var _this = this;
        this.debug = debug_1.default('local-machine-network:' + (options.id || ('[' + options.path + ']')));
        // Create the low-level network
        this.lowLevel = new low_level_1.LowLevelNetwork(options);
        this[connectEvent] = new atvik_1.Event(this);
        this[connectionEvent] = new atvik_1.Event(this);
        this[leaderEvent] = new atvik_1.Event(this);
        this[messageEvent] = new atvik_1.Event(this);
        /*
         * When this instance becomes the leader make sure to intercept
         * server socket errors and to emit the leader event.
         */
        this.lowLevel.onLeader(function (serverSocket) {
            // Listen to all socket errors to avoid unexpected exits
            serverSocket.on('error', function (err) {
                _this.debug('Caught error from server socket', err);
            });
            // If previously connected to a leader remove the reference
            _this.connection = undefined;
            // Emit the leader event
            _this[leaderEvent].emit();
        });
        /*
         * On incoming connections make sure to wrap them with encoding and
         * decoding.
         */
        this.lowLevel.onConnection(function (clientSocket) {
            // Create the wrapped socket
            var connection = new ObjectSocket({
                socket: clientSocket,
                debug: _this.debug,
                emitMessage: function (msg) { return _this[messageEvent].emit(msg); },
                send: _this.send.bind(_this)
            });
            _this[connectionEvent].emit(connection);
        });
        /*
         * When this instance is not the leader and it connects to something
         * save a reference to the connection.
         */
        this.lowLevel.onConnect(function (socket) {
            var connection = _this.connection = new ObjectSocket({
                socket: socket,
                debug: _this.debug,
                emitMessage: function (msg) { return _this[messageEvent].emit(msg); },
                send: _this.send.bind(_this)
            });
            _this[connectEvent].emit(connection);
        });
    }
    Object.defineProperty(ObjectNetwork.prototype, "onConnect", {
        get: function () {
            return this[connectEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ObjectNetwork.prototype, "onLeader", {
        get: function () {
            return this[leaderEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ObjectNetwork.prototype, "onConnection", {
        get: function () {
            return this[connectionEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ObjectNetwork.prototype, "onMessage", {
        get: function () {
            return this[messageEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Connect to the network.
     */
    ObjectNetwork.prototype.start = function () {
        return this.lowLevel.start();
    };
    /**
     * Disconnect from the network.
     */
    ObjectNetwork.prototype.stop = function () {
        return this.lowLevel.stop();
    };
    /**
     * Send a message to the current leader.
     */
    ObjectNetwork.prototype.send = function (message) {
        if (this.lowLevel.leader) {
            // This instance is the leader, emit the event locally
            this[messageEvent].emit({
                returnPath: this,
                data: message
            });
        }
        else if (this.connection) {
            // Connected to the leader send via the active socket
            this.connection.send(message);
        }
        else {
            this.debug('Unable to send message, not connected, message was:', message);
            throw new Error('Unable to send message, not connected');
        }
    };
    return ObjectNetwork;
}());
exports.ObjectNetwork = ObjectNetwork;
var ObjectSocket = /** @class */ (function () {
    function ObjectSocket(control) {
        var _this = this;
        this.control = control;
        this[disconnectEvent] = new atvik_1.Event(this);
        this[messageEvent] = new atvik_1.Event(this);
        // Intercept errors and output them as debug info
        control.socket.on('error', function (err) {
            control.debug('Error received:', err);
        });
        // Emit disconnection events
        control.socket.on('close', function () { return _this[disconnectEvent].emit(); });
        // Setup the decoder for incoming messages
        var decoder = msgpack_lite_1.default.createDecodeStream();
        var pipe = control.socket.pipe(decoder);
        // When data is received emit an event
        decoder.on('data', function (data) {
            control.debug('Received message:', data);
            var message = {
                returnPath: _this,
                data: data
            };
            control.emitMessage(message);
            _this[messageEvent].emit(message);
        });
        // Catch errors on pipe and decoder
        decoder.on('error', function (err) { return control.debug('Error from decoder', err); });
        pipe.on('error', function (err) { return control.debug('Error from pipe', err); });
    }
    Object.defineProperty(ObjectSocket.prototype, "onDisconnect", {
        get: function () {
            return this[disconnectEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ObjectSocket.prototype, "onMessage", {
        get: function () {
            return this[messageEvent].subscribable;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Send a message to this instance.
     *
     * @param {*} message
     */
    ObjectSocket.prototype.send = function (message) {
        var data = msgpack_lite_1.default.encode(message);
        try {
            this.control.debug('Sending message to leader:', message);
            this.control.socket.write(data);
        }
        catch (err) {
            this.control.debug('Could not write, got an error:', err);
        }
    };
    return ObjectSocket;
}());
exports.ObjectSocket = ObjectSocket;
//# sourceMappingURL=object.js.map