/// <reference types="node" />
import debug from 'debug';
import { Socket } from 'net';
declare const connectEvent: unique symbol;
declare const disconnectEvent: unique symbol;
declare const leaderEvent: unique symbol;
declare const connectionEvent: unique symbol;
declare const messageEvent: unique symbol;
/**
 * Options for creating an object network.
 */
export interface ObjectNetworkOptions {
    /**
     * Optional id used for debug messages.
     */
    id?: string;
    /**
     * Path at which the network should create its socket.
     */
    path: string;
}
/**
 * Interface for things that can send an object message.
 */
export interface ObjectSender {
    send(message: any): void;
}
/**
 * Message that can be received via the object network.
 */
export interface ObjectMessage {
    /**
     * Return path of the message, can be used to send a message to whoever
     * sent this message.
     */
    returnPath: ObjectSender;
    /**
     * Data carried by message.
     */
    data: any;
}
/**
 * Adapter over a low-level network for those cases where clients want to
 * exchange objects with the leader. This implementation will use MessagePack
 * to encode and decode objects.
 */
export declare class ObjectNetwork {
    private debug;
    private lowLevel;
    private connection?;
    private [connectEvent];
    private [connectionEvent];
    private [leaderEvent];
    private [messageEvent];
    /**
     * Create a new object network.
     */
    constructor(options: ObjectNetworkOptions);
    readonly onConnect: import("atvik").Subscribable<this, [ObjectSocket]>;
    readonly onLeader: import("atvik").Subscribable<this, []>;
    readonly onConnection: import("atvik").Subscribable<this, [ObjectSocket]>;
    readonly onMessage: import("atvik").Subscribable<this, [ObjectMessage]>;
    /**
     * Connect to the network.
     */
    start(): Promise<void>;
    /**
     * Disconnect from the network.
     */
    stop(): Promise<void>;
    /**
     * Send a message to the current leader.
     */
    send(message: any): void;
}
interface ObjectSocketControl {
    send(message: any): void;
    emitMessage(message: ObjectMessage): void;
    debug: debug.Debugger;
    socket: Socket;
}
export declare class ObjectSocket {
    private control;
    private [disconnectEvent];
    private [messageEvent];
    constructor(control: ObjectSocketControl);
    readonly onDisconnect: import("atvik").Subscribable<this, []>;
    readonly onMessage: import("atvik").Subscribable<this, [ObjectMessage]>;
    /**
     * Send a message to this instance.
     *
     * @param {*} message
     */
    send(message: any): void;
}
export {};
